package org.btcollider.cnc.keysrv.impl.multistage;

import org.btcollider.cnc.dto.KeyRange;
import org.btcollider.cnc.keysrv.KeyServer;
import org.btcollider.cnc.keysrv.impl.keytree.KeyTreeServer;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

/**
 * Not entirely random KeyServer that handles very large trees in multiple stages to reduce memory
 * and processing time consumption
 *
 * @author armin
 *
 */
public class MultiStagedServer implements KeyServer {
  final Logger log = LoggerFactory.getLogger(MultiStagedServer.class);
  private static final int PARTITION_DEPTH = 15;

  private KeyTreeServer primKTS;
  private KeyTreeServer secKTS;
  private KeyRange partition;
  private KeyRangeBuffer keyRangeBuffer;
  private int remainingDepth;
  private int subIndex;
  private long maxWorkSpan;

  public MultiStagedServer(int index, int depth, long maxWorkSpan, long pruneKey) {
    this.remainingDepth = depth - PARTITION_DEPTH;
    this.subIndex = index - PARTITION_DEPTH;
    this.maxWorkSpan = maxWorkSpan;

    // Build a reasonably sized KeyTreeServer with disabled KTConcierge service
    // This represents the first partitioning of the key space
    this.primKTS = new KeyTreeServer(index, PARTITION_DEPTH, 0, pruneKey);
    // Get a random starting range from the primary KTS, this range is still too huge
    this.partition = primKTS.getRange();
    // Start a secondary KTS to partition remaining key space in currentRange
    // This represents the actual work-horse KeyTreeServer for the current partition with
    // enabled KTConcierge service.
    // Needs subindex+1 because starting bit always set to 1
    long subPrune = ((long) 1 << subIndex + 1) | ((((long) 1 << subIndex + 1) - 1) & pruneKey);
    this.secKTS = new KeyTreeServer(subIndex + 1, remainingDepth, maxWorkSpan, subPrune);

    this.keyRangeBuffer = new KeyRangeBuffer(maxWorkSpan);
  }

  @Override
  public KeyRange getRange() {
    if (partition == null)
      return null; // We are finished searching the whole space

    KeyRange kr = null;

    if ((kr = keyRangeBuffer.getRange()) == null) { // Get retracted KR from buffer if one exists
      kr = secKTS.getRange(); // else: get a random range from the subspace
    }

    if (kr == null) { // sub-range finished
      log.info("Subrange {} finished. Opening new partition.", partition);
      primKTS.setSearched(partition);
      partition = primKTS.getRange();
      secKTS = new KeyTreeServer(subIndex + 1, remainingDepth, maxWorkSpan, 0);
      return getRange();
    }

    // strip out starting bit of subrange
    long krStart = kr.getStart() & ~(1 << subIndex + 1);
    // extract set tail bits
    long krEnd = kr.getStart() ^ kr.getEnd();

    // Combine partition with sub-range
    krStart = partition.getStart() | krStart;
    krEnd = krStart | krEnd;

    return new KeyRange(krStart, krEnd);
  }

  @Override
  public void setInWork(KeyRange keyRange) {
    keyRangeBuffer.setInWork(keyRange);

    // strip out partition bits -> get subspace KeyRange only
    long mask = (long) Math.pow(2, (subIndex + 1)) - 1;
    // add leading bit
    long lb = (long) Math.pow(2, (subIndex + 1));

    KeyRange secKR = new KeyRange(keyRange.getStart() & mask | lb, keyRange.getEnd() & mask | lb);
    secKTS.setInWork(secKR);
  }

  @Override
  public void setSearched(KeyRange keyRange) {
    keyRangeBuffer.setSearched(keyRange);

    // strip out partition bits -> get subspace KeyRange only
    long mask = (long) Math.pow(2, (subIndex + 1)) - 1;
    // add leading bit
    long lb = (long) Math.pow(2, (subIndex + 1));

    KeyRange secKR = new KeyRange(keyRange.getStart() & mask | lb, keyRange.getEnd() & mask | lb);
    secKTS.setSearched(secKR);
  }

}
